<img src="http://dalkescientific.com/logos/chemfp_1.4_logo.png">
<h3>{{ _('Links') }}</h3>
<ul>
  <li><a href="http://dalkescientific.com/releases/chemfp-1.4.tar.gz">Download chemfp 1.4</a></li>
  <li><a href="http://chemfp.com/">chemfp home page</a></li>
  <li><a href="https://readthedocs.org/projects/chemfp/">chemfp at Read The Docs</a></li>
  <li>Download documentation as:</li>
  <ul>
    <li><a href="https://readthedocs.org/projects/chemfp/downloads/pdf/chemfp-1.4/">PDF</a></li>
    <li><a href="https://readthedocs.org/projects/chemfp/downloads/htmlzip/chemfp-1.4/">HTMLZip</li>
  </ul>
</ul>
